@@ -5,7 +5,9 @@ use bitvec_helpers::{bitvec_reader::BitVecReader, bitvec_writer::BitVecWriter};
 use serde::{Deserialize, Serialize};
 
 use super::dovi_rpu::DoviRpu;
-use super::extension_metadata::blocks::{ExtMetadataBlock, ExtMetadataBlockLevel11};
+use super::extension_metadata::blocks::{
+    ExtMetadataBlock, ExtMetadataBlockLevel11, ExtMetadataBlockLevel9,
+};
 use super::extension_metadata::*;
 use super::generate::GenerateConfig;
 use super::profiles::profile81::Profile81;
@@ -363,6 +365,7 @@ impl VdrDmData {
             ExtMetadataBlock::Level254(_) => {
                 bail!("Cannot replace automatically generated Level254 block")
             }
+            ExtMetadataBlock::Level255(_) => self.replace_metadata_level(block),
             ExtMetadataBlock::Reserved(_) => bail!("Cannot replace specific reserved block"),
         }
     }
@@ -443,7 +446,12 @@ impl VdrDmData {
             }
             CmVersion::V40 => {
                 vdr_dm_data.cmv29_metadata = Some(DmData::V29(CmV29DmData::default()));
-                vdr_dm_data.cmv40_metadata = Some(DmData::V40(CmV40DmData::new_with_l254()));
+
+                vdr_dm_data.cmv40_metadata = if let Some(level254) = &config.level254 {
+                    Some(DmData::V40(CmV40DmData::new_with_custom_l254(level254)))
+                } else {
+                    Some(DmData::V40(CmV40DmData::new_with_l254_402()))
+                }
             }
         }
 
@@ -456,17 +464,22 @@ impl VdrDmData {
     pub fn set_static_metadata(&mut self, config: &GenerateConfig) -> Result<()> {
         self.replace_metadata_block(ExtMetadataBlock::Level5(config.level5.clone()))?;
         self.replace_metadata_block(ExtMetadataBlock::Level6(config.level6.clone()))?;
+
+        // Default to inserting both L9 (required) and L11 metadata
+        self.replace_metadata_block(ExtMetadataBlock::Level9(
+            ExtMetadataBlockLevel9::default_dci_p3(),
+        ))?;
         self.replace_metadata_block(ExtMetadataBlock::Level11(
             ExtMetadataBlockLevel11::default_reference_cinema(),
         ))?;
 
         if !config.default_metadata_blocks.is_empty() {
-            let level_block_list: &[u8] = &[5, 6, 254];
+            const LEVEL_BLOCK_LIST: &[u8] = &[5, 6, 254];
 
             let allowed_default_blocks = config
                 .default_metadata_blocks
                 .iter()
-                .filter(|block| !level_block_list.contains(&block.level()));
+                .filter(|block| !LEVEL_BLOCK_LIST.contains(&block.level()));
 
             for block in allowed_default_blocks {
                 self.replace_metadata_block(block.clone())?;
